/*
 * This file is part of SionTower.
 *
 * 
 * David Saltares Márquez (C) 2011
 * <david.saltares@gmail.com>
 *
 * 
 * SionTower examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * SionTower examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SionTower examples.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 *  @file shape.cpp
 *  @author David Saltares Márquez
 *  @date 25-01-2011
 */

#include <iostream>
#include <algorithm>

#include <boost/bind.hpp>

#include "shape.h"

using std::cerr;
using std::cout;
using std::endl;

Shape::CollisionDispatchTable Shape::_collisionDispatcher;

Shape::Shape(const Ogre::String& name): _name(name) {
    cout << "Shape::Shape()" << endl;
}

Shape::~Shape() {
    cout << "Shape::~Shape()" << endl;
}

const Ogre::String& Shape::getName() const {
    return _name;
}

void Shape::configureCollisionDispatching() {
    // Completamos la tabla de chequeos de colisión
    _collisionDispatcher[SPHERE][SPHERE] = boost::bind(&Shape::getCollisionSphereSphere, _1, _2);
    _collisionDispatcher[AABB][AABB] = boost::bind(&Shape::getCollisionAABBAABB, _1, _2);
    _collisionDispatcher[PLANE][PLANE] = boost::bind(&Shape::getCollisionPlanePlane, _1, _2);

    _collisionDispatcher[AABB][SPHERE] = boost::bind(&Shape::getCollisionSphereAABB, _1, _2);
    _collisionDispatcher[AABB][PLANE] = boost::bind(&Shape::getCollisionPlaneAABB, _1, _2);

    _collisionDispatcher[SPHERE][AABB] = boost::bind(&Shape::getCollisionSphereAABB, _1, _2);
    _collisionDispatcher[SPHERE][PLANE] = boost::bind(&Shape::getCollisionPlaneSphere, _1, _2);

    _collisionDispatcher[PLANE][AABB] = boost::bind(&Shape::getCollisionPlaneAABB, _1, _2);
    _collisionDispatcher[PLANE][SPHERE] = boost::bind(&Shape::getCollisionPlaneSphere, _1, _2);
}

bool Shape::getCollision(Shape* shapeA, Shape* shapeB) {
    // Comprobamos si la forma A está registrada
    CollisionDispatchTable::iterator itA = _collisionDispatcher.find(shapeA->getType());
    if (itA == _collisionDispatcher.end()) {
        cout << "Shape::getCollision(): no existe el tipo " << shapeA->getType() << endl;
        return 0;
    }
    
    // Comprobamos que la forma B está registrada
    CollisionDispatchTable::iterator itB = _collisionDispatcher.find(shapeB->getType());
    if (itB == _collisionDispatcher.end()) {
        cout << "Shape::getCollision(): no existe el tipo " << shapeB->getType() << endl;
        return 0;
    }

    // Comprobamos que hay un método de comprobación del tipo A - B
    boost::unordered_map<int, CollisionCheckFunction>::iterator itC = _collisionDispatcher[shapeA->getType()].find(shapeB->getType());
    if (itC == itA->second.end()) {
        cout << "Shape::getCollision(): no existe un método de comprobación entre" << shapeA->getType() << " y " << shapeB->getType() << endl;
        return 0;
    }

    // Llamamos al método de comprobación
    return itC->second(shapeA, shapeB);
}


bool Shape::getCollisionSphereSphere(Shape* shapeA, Shape* shapeB) {
    // Hacemos la conversión (estamos seguros de que son esferas)
    Sphere* sphereA = static_cast<Sphere*>(shapeA);
    Sphere* sphereB = static_cast<Sphere*>(shapeB);

    // Hacemos el test
    Ogre::Vector3 s = sphereA->getCenter() - sphereB->getCenter();
    Ogre::Real totalRadius = sphereA->getRadius() + sphereB->getRadius();

    return (s.squaredLength() <= totalRadius * totalRadius);
}


bool Shape::getCollisionAABBAABB(Shape* shapeA, Shape* shapeB) {
    // Hacemos la conversión (estamos seguros de que son AABBs)
    AxisAlignedBox* aabb1= static_cast<AxisAlignedBox*>(shapeA);
    AxisAlignedBox* aabb2 = static_cast<AxisAlignedBox*>(shapeB);

    // Hacemos el test
    return (aabb1->getMaxPos().x > aabb2->getMinPos().x &&
            aabb1->getMinPos().x < aabb2->getMaxPos().x &&
            aabb1->getMaxPos().y > aabb2->getMinPos().y &&
            aabb1->getMinPos().y < aabb2->getMaxPos().y &&
            aabb1->getMaxPos().z > aabb2->getMinPos().z &&
            aabb1->getMinPos().z < aabb2->getMaxPos().z);
}

bool Shape::getCollisionSphereAABB(Shape* shapeA, Shape* shapeB) {
    // Hacemos la conversión (estamos seguros de que A es Sphere y B es AABB)
    Sphere* sphere;
    AxisAlignedBox* aabb;
    if (shapeA->getType() == SPHERE) {
        sphere = static_cast<Sphere*>(shapeA);
        aabb = static_cast<AxisAlignedBox*>(shapeB);
    } else {
        sphere = static_cast<Sphere*>(shapeB);
        aabb = static_cast<AxisAlignedBox*>(shapeA);
    }

    // Hacemos el test
    Ogre::Real s = 0;
    Ogre::Real d = 0;
    Ogre::Vector3 center = sphere->getCenter();
    Ogre::Vector3 minPos = aabb->getMinPos();
    Ogre::Vector3 maxPos = aabb->getMaxPos();

    // Comprobamos si el centro de la esfera está dentro del AABB
    bool centerInsideAABB = (center.x <= maxPos.x &&
                             center.x >= minPos.x &&
                             center.y <= maxPos.y &&
                             center.y >= minPos.y &&
                             center.z <= maxPos.z &&
                             center.z >= minPos.z);

    if (centerInsideAABB)
        return true;

    // Comprobamos si la esfera y el AABB se intersectan
    for (int i = 0; i < 3; ++i) {
        if (sphere->getCenter()[i] < aabb->getMinPos()[i]) {
            s = sphere->getCenter()[i] - aabb->getMinPos()[i];
            d += s * s;
        } else if (sphere->getCenter()[i] > aabb->getMaxPos()[i]) {
            s = sphere->getCenter()[i] - aabb->getMaxPos()[i];
            d += s * s;
        }
    }

    return (d <= sphere->getRadius() * sphere->getRadius());
}

bool Shape::getCollisionPlanePlane(Shape* shapeA, Shape* shapeB) {
    // Hacemos la conversión (estamos seguros de que son Planes)
    Plane* planeA = static_cast<Plane*>(shapeA);
    Plane* planeB = static_cast<Plane*>(shapeB);

    // Hacemos el test
    Ogre::Vector3 normalA = planeA->getNormal().normalisedCopy();
    Ogre::Vector3 normalB = planeB->getNormal().normalisedCopy();
    return normalA.dotProduct(normalB) != 1;
}

bool Shape::getCollisionPlaneSphere(Shape* shapeA, Shape* shapeB) {
    // Hacemos la conversión (estamos seguros de que A es Plane y B es Sphere)
    Plane* plane;
    Sphere* sphere;
    if (shapeA->getType() == PLANE) {
        plane = static_cast<Plane*>(shapeA);
        sphere = static_cast<Sphere*>(shapeB);
    } else {
        plane = static_cast<Plane*>(shapeB);
        sphere = static_cast<Sphere*>(shapeA);
    }

    // Hacemos el test
    
    // Distancia del centro de la esfera al plano
    Ogre::Vector3 v = sphere->getCenter() - plane->getPosition();
    Ogre::Vector3 n = plane->getNormal().normalisedCopy();
    Ogre::Real d = abs(n.dotProduct(v));

    // Si d <= radio, hay colisión
    return d <= sphere->getRadius();
}

bool Shape::getCollisionPlaneAABB(Shape* shapeA, Shape* shapeB) {
    // Hacemos la conversión (estamos seguros de que A es Plane y B es AABB)
    Plane* plane;
    AxisAlignedBox* aabb;
    if (shapeA->getType() == PLANE) {
        plane = static_cast<Plane*>(shapeA);
        aabb = static_cast<AxisAlignedBox*>(shapeB);
    } else {
        plane = static_cast<Plane*>(shapeB);
        aabb = static_cast<AxisAlignedBox*>(shapeA);
    }


    // Hacemos el test
    Ogre::Vector3 p;
    Ogre::Vector3 n;

    for (int i = 0; i < 3; ++i) {
        if (plane->getNormal()[i] >= 0) {
            p[i] = aabb->getMaxPos()[i];
            n[i] = aabb->getMinPos()[i];
        } else {
            p[i] = aabb->getMaxPos()[i];
            n[i] = aabb->getMinPos()[i];
        }
    }

    // Si p está en un lado diferente del plano que n, hay intersección
    Ogre::Real d1 = plane->getNormal().dotProduct(p - plane->getPosition());
    Ogre::Real d2 = plane->getNormal().dotProduct(n - plane->getPosition());

    return ((d1 <= 0 && d2 >= 0) || (d1 >= 0 && d2 <= 0));
}


/*
 *  SPHERE
 */


Sphere::Sphere(const Ogre::String& name,
               const Ogre::Vector3& center,
               Ogre::Real radius): Shape(name), _center(center), _radius(radius) {
    cout << "Sphere::Sphere()" << endl;
}

Sphere::~Sphere() {
    cout << "Sphere::~Sphere()" << endl;
}

int Sphere::getType() const {
    return SPHERE;
}

void Sphere::applyTransform(Shape* localShape,
                            const Ogre::Vector3& traslation,
                            const Ogre::Vector3& scale,
                            const Ogre::Quaternion& orientation) {
    // localShape debería ser una esfera
    Sphere* localSphere = static_cast<Sphere*>(localShape);

    // Aplicamos traslación y escala (no es necesario rotar la esfera)
    _center = localSphere->_center + traslation;
    // No podemos escalar una esfera teniendo en cuenta x, y, z porque no
    // podemos deformarla.
    _radius = localSphere->_radius * scale.x;
}
        
Shape* Sphere::getTransformedCopy(const Ogre::Vector3& traslation,
                                  const Ogre::Vector3& scale,
                                  const Ogre::Quaternion& orientation) {
    // Creamos la nueva esfera
    Sphere* sphere = new Sphere(_name, _center, _radius);

    // Aplicamos la transformacion
    sphere->applyTransform(this, traslation, scale, orientation);

    return sphere;
}

Ogre::Real Sphere::getRadius() const {
    return _radius;
}
        
void Sphere::setRadius(Ogre::Real radius) {
    _radius = radius;
}

const Ogre::Vector3& Sphere::getCenter() const {
    return _center;
}

void Sphere::setCenter(const Ogre::Vector3& center) {
    _center = center;    
}





/*
 *  AXISALIGNEDBOX
 */


AxisAlignedBox::AxisAlignedBox(const Ogre::String& name,
                               const Ogre::Vector3& minPos,
                               const Ogre::Vector3& maxPos): Shape(name), _minPos(minPos), _maxPos(maxPos) {
    cout << "AxisAlignedBox::AxisAlignedBox()" << endl;
}

AxisAlignedBox::~AxisAlignedBox() {
    cout << "AxisAlignedBox::~AxisAlignedBox()" << endl;
}

int AxisAlignedBox::getType() const {
    return AABB;
}
        
void AxisAlignedBox::applyTransform(Shape* localShape,
                                    const Ogre::Vector3& traslation,
                                    const Ogre::Vector3& scale,
                                    const Ogre::Quaternion& orientation) {
    // localShape debería ser AxisAlignedBox
    AxisAlignedBox* localAABB = static_cast<AxisAlignedBox*>(localShape);

    // Aplicamos traslación
    _minPos = localAABB->_minPos + traslation;
    _maxPos = localAABB->_maxPos + traslation;

    // Aplicamos la escala
    _minPos *= scale;
    _maxPos *= scale;

    // No podemos rotar un AABB
}

Shape* AxisAlignedBox::getTransformedCopy(const Ogre::Vector3& traslation,
                                          const Ogre::Vector3& scale,
                                          const Ogre::Quaternion& orientation) {
    // Creamos el nuevo AABB
    AxisAlignedBox* aabb = new AxisAlignedBox(_name, _minPos, _maxPos);

    // Aplicamos la transformacion
    aabb->applyTransform(this, traslation, scale, orientation);

    return aabb;
}

const Ogre::Vector3& AxisAlignedBox::getMinPos() const {
    return _minPos;
}

void AxisAlignedBox::setMinPos(const Ogre::Vector3& minPos) {
    _minPos = minPos;
}

const Ogre::Vector3& AxisAlignedBox::getMaxPos() const {
    return _maxPos;
}

void AxisAlignedBox::setMaxPos(const Ogre::Vector3& maxPos) {
    _maxPos = maxPos;
}



/*
 *  PLANE
 */


Plane::Plane(const Ogre::String& name,
             const Ogre::Vector3& position,
             const Ogre::Vector3& normal): Shape(name), _position(position), _normal(normal) {
    cout << "Plane::Plane()" << endl;
}

Plane::~Plane() {
    cout << "Plane::~Plane()" << endl;
}

int Plane::getType() const {
    return PLANE;
}
        
void Plane::applyTransform(Shape* localShape,
                           const Ogre::Vector3& traslation,
                           const Ogre::Vector3& scale,
                           const Ogre::Quaternion& orientation) {
    
    // localShape debería ser Plane
    Plane* plane = static_cast<Plane*>(localShape);

    // Aplicamos traslacion
    _position = plane->_position + traslation;

    // No tiene sentido aplicar escala
    
    // Aplicamos la rotación a la normal
    _normal = orientation * plane->_normal;
}

Shape* Plane::getTransformedCopy(const Ogre::Vector3& traslation,
                                 const Ogre::Vector3& scale,
                                 const Ogre::Quaternion& orientation) {
    // Creamos un nuevo plano
    Plane* plane = new Plane(_name, _position, _normal);

    // Aplicamos la transformacion
    plane->applyTransform(this, traslation, scale, orientation);

    return plane;
}


const Ogre::Vector3& Plane::getPosition() const {
    return _position;
}

void Plane::setPosition(const Ogre::Vector3& position) {
    _position = position;
}

const Ogre::Vector3& Plane::getNormal() const {
    return _normal;
}

void Plane::setNormal(const Ogre::Vector3& normal) {
    _normal = normal;
}

